/** This file is part of GeodeMAPS, a program for computing the geodesic distance between phylogenetic trees.
    Copyright (C) 2008  Megan Owen

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


package distanceAlg1;

import java.util.*;

// XXX get rid of dist variable???
public class Geodesic {
	private RatioSequence rs;
//	private double dist;
	private Vector<PhyloTreeEdge> commonEdges;   // stores the common edges where their length is the difference in lengths of that edge between the two trees
	private double leafContributionSquared = 0;
	
	// constructor
/*	public Geodesic(double dist, RatioSequence rs) {
		this.dist = dist;
		this.rs = rs;
	}
*/	
	public Geodesic(RatioSequence rs) {
		this.rs = rs;
		//		this.rs = rs.getAscRSWithMinDist();
		commonEdges = new Vector<PhyloTreeEdge>();
//		dist = this.rs.getDistance();
	}
	
	public Geodesic(RatioSequence rs, Vector<PhyloTreeEdge> cEdges) {
		this.rs = rs;
		//		this.rs = rs.getAscRSWithMinDist();
		commonEdges = cEdges;
	}
	
	public Geodesic(RatioSequence rs, Vector<PhyloTreeEdge> cEdges, double leafContributionSquared) {
		this.rs = rs;
		//		this.rs = rs.getAscRSWithMinDist();
		commonEdges = cEdges;
		this.leafContributionSquared = leafContributionSquared;
	}
	
	public RatioSequence getRS() {
		return rs;
	}
	
	public void setRS(RatioSequence rs) {
		this.rs = rs;
//		dist = rs.getAscRSWithMinDist().getDistance();
	}
	public double getDist() {
		double commonEdgeDistSquared = 0;
		for(int i = 0; i < commonEdges.size(); i++) {
			commonEdgeDistSquared = commonEdgeDistSquared + Math.pow(commonEdges.get(i).getLength(),2);
		}
/*		if (TreeDistance.verbose > 0) {
			System.out.println("rs.getAscRSWithMinDist().getDistance() is " + rs.getAscRSWithMinDist().getDistance() + "; commonEdgeDistSquared is " +commonEdgeDistSquared + "; and leafContributionSquared is " + leafContributionSquared);
		} */
		return Math.sqrt(Math.pow(rs.getAscRSWithMinDist().getDistance(), 2) + commonEdgeDistSquared + leafContributionSquared);
	}
/*	public void setDist(double dist) {
		this.dist = dist;
	}
*/	
	/** Adds edge to list of common edges.  Assumes the length is the change in this common edge
	 * between the two trees.
	 */
	public void addCommonEdge(PhyloTreeEdge e) {
		commonEdges.add(e);
	}
	
	/**  Displays the geodesic in user-friendly form.
	 * 
	 * @return
	 */
	public String toStringVerboseOld(PhyloTree t1, PhyloTree t2) {
		Vector<PhyloTreeEdge> commonEdges = this.getCommonEdges();
		Boolean cEdge = false;
		
		// display T1 only splits
		String toDisplay = "\nSplits only in T1:\n";
		for (int i = 0; i < t1.getEdges().size(); i++) {
			cEdge = false;
			for (int j = 0; j < commonEdges.size(); j++) {
				if (commonEdges.get(j).sameBipartition(t1.getEdge(i))) {
					cEdge = true;
					break;
				}
			}
			if (!cEdge) {
				toDisplay = toDisplay + t1.getEdge(i).toStringVerbose(t1.getLeaf2NumMap()) + "\n" + t1.getEdge(i) + "\n";
			}
		}
		
		//display T2 only splits
		toDisplay = toDisplay + "\n\nSplits only in T2:\n";
		for (int i = 0; i < t2.getEdges().size(); i++) {
			if (! commonEdges.contains(t2.getEdge(i)) ) {
				// if this is not a common edge, display
//				toDisplay = toDisplay + t2.getEdge(i).toStringVerbose(t2.getLeaf2NumMap()) + "\n";
				toDisplay = toDisplay + t2.getEdge(i).toStringVerbose(t2.getLeaf2NumMap()) + "\n" + t2.getEdge(i) + "\n";
			}
		}
		
		// display common splits
		toDisplay = toDisplay + "\n\nCommon splits:\n";
		for (int i = 0; i < commonEdges.size(); i++) {
//			toDisplay = toDisplay + commonEdges.get(i).toStringVerbose(t1.getLeaf2NumMap()) + "\n";
			toDisplay = toDisplay + commonEdges.get(i).toStringVerbose(t1.getLeaf2NumMap()) + "\n" + commonEdges.get(i) + "\n";
		}
		
		return toDisplay;
	}
	
	public Geodesic clone() { 
		return new Geodesic(rs.clone(), TreeDistance.myVectorClonePhyloTreeEdge(commonEdges));
	}
	
	public String toString() {
		return "" + getDist() + "; " + rs.getAscRSWithMinDist();
//		return "" + getDist() + "; " + rs;
	}

	public Vector<PhyloTreeEdge> getCommonEdges() {
		return commonEdges;
	}

	public void setCommonEdges(Vector<PhyloTreeEdge> commonEdges) {
		this.commonEdges = commonEdges;
	}
	
	/** Returns the geodesic with the ratio sequence (and ratios) reversed.
	 * 
	 * @return
	 */
	public Geodesic reverse() {
		return new Geodesic(rs.reverse(), commonEdges, leafContributionSquared);
	}

	public double getLeafContributionSquared() {
		return leafContributionSquared;
	}

	public void setLeafContributionSquared(double leafContributionSquared) {
		this.leafContributionSquared = leafContributionSquared;
	}
}
